#!/usr/bin/perl -w
# 
#    SaintMap v2.0 - Visual configuration tool for NetSaint
#
#    Copyright (C) 2000 David Kmoch <David.Kmoch@vslib.cz>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.


##################################################
#
# Path for directory with images 
#
  my $imgpath="./images/";
#
##################################################
#
# Background image path. Use only GIF images.
# Set to empty string if you do not want any
# background image
#
# my $bg_image="tenmap.gif";
  my $bg_image="";
#
#  
##################################################
#       NO CHANGES NEEDED BELOW                  # 
##################################################

use strict;
use Tk;

my %object;
my %objprop;
my @images;
my($main,$menubar,$canvas,$dlg);
my($canvw,$canvh)=(4096,2048);	# Canvas width & height
my($canvx,$canvy)=(1024,768);	# Viewable canvas area
my($xpos,$ypos)=(0,0);

&init_gui;

MainLoop;

############################ SUBS #################################
sub init_gui {	# Initialize main widget, menubar and canvas
  $main=MainWindow->new(-title=>'SaintMap');
  $main->resizable(0,0);
  $main->protocol('WM_DELETE_WINDOW',\&quit_it);
  &create_menubar;
  &create_canvas;
  &bgimage;
}  
############
sub create_menubar {	# Create menubar, specify callbacks
  my($menuframe,$coordframe,$gridb,$xlabel,$ylabel,$xvalue,$yvalue);
  $menuframe=$main->Frame(qw/-relief raised -borderwidth 2/)
                     ->pack(qw/-side top -anchor n -expand 1 -fill x/);

  $menubar=$menuframe->Menubutton(qw/-text File -underline 0 -menuitems/ =>
   [
    [Button => 'Open', -command =>
      [\&file_dialog,"Open file","Select file to read","hosts.cfg",\&open_hosts]],
    [Button => 'Save', -command =>
      [\&file_dialog,"Save file","Save to file:","", \&save_state],
                       -state=>'disabled'],
    [Button => 'Quit', -command => \&quit_it]
   ])->pack(-side => 'left');
   
  $gridb=$menuframe->Button(-text=>'Grid',
                            -relief=>'flat',
			    -command=>\&mygrid,
			   )
                   ->pack(-side,'left');
  $coordframe=$menuframe->Frame(-relief,'groove',-borderwidth,2)->pack(-side=>'right');
  $xlabel=$coordframe->Label(-text => ' X ')->pack(-side => 'left');
  $xvalue=$coordframe->Label(-background=>'white',
                             -takefocus=>0,
			     -textvariable=>\$xpos,
			     -width=>4,
			     )
                     ->pack(-side => 'left');
  $ylabel=$coordframe->Label(-text => ' Y ')->pack(-side => 'left');
  $yvalue=$coordframe->Label(-background=>'white',
                             -takefocus=>0,
			     -textvariable=>\$ypos,
			     -width=>4,
			     )
                     ->pack(-side => 'left');
}
############
sub create_canvas {	# Create canvas, bind specific actions
  my $c=$main->Scrolled("Canvas",-width => $canvx,
                                 -height => $canvy,
				 -scrollregion => [0,0,$canvw,$canvh],
				 -bg=>'gray70',
				 -scrollbars=>'osoe',
				 )
             ->pack(-fill => 'both', -expand => 1);
  $canvas=$c->Subwidget("canvas");
  $canvas->Tk::bind("<1>", [ \&get_pos, Ev('x'), Ev('y')]);
  $canvas->Tk::bind("<B1-Motion>", [ \&get_pos, Ev('x'), Ev('y')]);
  $canvas->bind("moveable","<1>", [ \&drag_start, Ev('x'), Ev('y') ]);
  $canvas->bind("moveable","<B1-Motion>", [ \&drag_it, Ev('x'), Ev('y') ]);
  $canvas->bind("moveable","<Any-Enter>", [ \&highlight]);
  $canvas->bind("moveable","<Any-Leave>", [ \&highlight ]);
  $canvas->bind("moveable","<3>", [ \&change_type]);
}  
############
sub bgimage { # If defined, show background image
  if(defined $bg_image) {
    my $img=$main->Photo(-file=>$bg_image);
    $canvas->createImage(0,0,-image=>$img,-anchor=>'nw',-tags=>["bgimg"]);
  }
}  
############
sub mygrid { # Show/hide grid
  my($i);
  my $step=50;
  if($canvas->find("withtag","grid")) {
    $canvas->delete("grid");
  } else {
    for ($i=$step;$i<$canvw;$i+=$step) {
      $canvas->createLine($i,0,$i,$canvh,-fill=>'gray60',-tags,["grid"]);
    }
    for ($i=$step;$i<$canvh;$i+=$step) {
      $canvas->createLine(0,$i,$canvw,$i,-fill=>'gray60',-tags,["grid"]);
    }
    $canvas->lower("grid","all");
    if(defined $bg_image) {$canvas->raise("grid","bgimg")}
  }
}  
############
sub get_pos {	# Callback for position indicators
  my ($obj,$x,$y)=@_;
  $xpos=$canvas->canvasx($x);
  $ypos=$canvas->canvasy($y);
}
############
sub get_geometry {  # Returns geometry string for toplevel widget's position
  my($x,$y); 
  $x=($main->rootx)+(100);
  $y=($main->rooty)+(100);
  return("+$x+$y");
}
############
sub highlight {	# Highlight object's label under the mouse cursor
  my($obj)=@_;
  my($e,$name,$action);
  $e=$obj->XEvent; 
  $name=($canvas->gettags($canvas->find("withtag","current")))[1];
  if($e->T eq 'EnterNotify') {
    $canvas->itemconfigure($object{$name}->{LabelID},-fill => 'Red');
  } else {
    $canvas->itemconfigure($object{$name}->{LabelID},-fill => 'Black');
    }
}  
############
sub drag_start {	# Callback for mouse button 1 click
  my ($obj,$x,$y) = @_;
  $objprop{x}=$x;
  $objprop{y}=$y;
  $objprop{obj}=$canvas->find("withtag","current");
}
############
sub drag_it {	# Callback for mouse button 1 motion -actually moves object, its label and lines
  my ($obj,$x,$y) = @_;
  my($dx,$dy,$name,$l1,$r1,$t1,$b1,$l2,$r2,$t2,$b2);
  $dx=($x-$objprop{x});
  $dy=($y-$objprop{y});
  $canvas->move($objprop{obj},$dx,$dy);
  $objprop{x}=$x;
  $objprop{y}=$y;
  $name=($canvas->gettags($objprop{obj}))[1];
  $canvas->move($object{$name}->{LabelID},$dx,$dy);
  ($object{$name}{Coord}{x},$object{$name}{Coord}{y})=$canvas->coords($objprop{obj});
  for my $line (keys %{$object{$name}{LineID}}) {
    ($l1,$r1,$t1,$b1)=$canvas->bbox($objprop{obj});
    ($l2,$r2,$t2,$b2)=$canvas->bbox($object{$object{$name}{LineID}{$line}}{"ID"});
    $canvas->coords($line,$l1+(int(($t1-$l1)/2)),$r1+(int(($b1-$r1)/2)),$l2+(int(($t2-$l2)/2)),$r2+(int(($b2-$r2)/2)));
  }
}
############
sub file_dialog {	# Show toplevel dialog, use supplied function as callback for OK button
  my($title,$text,$defval,$proc)=@_;
  my($label,$entry,$df,$okb,$kob);
  $dlg=$main->Toplevel(-title=>$title);
  $dlg->geometry(&get_geometry);
  $dlg->transient($main);
  $dlg->grab;
  $label=$dlg->Label(-text=>$text)->pack(-padx,5,-pady,5);
  $entry=$dlg->Entry(-relief=>'sunken',
                     -bg=>'white',
		     -width=>20,
		     -textvariable=>\$defval,
		    )
             ->pack(-pady=>5,-padx=>5);
  $entry->focus;
  $entry->selectionRange(0,'end');
  $entry->bind("<Return>",sub{$okb->invoke});
  $df=$dlg->Frame()->pack(-fill=>'x',-expand=>1);
  $okb=$df->Button(-text=>'Ok',
                   -command=>sub {&$proc($defval)},
		  )
          ->pack(-padx=>5,-side=>'left');
  $kob=$df->Button(-text=>'Cancel',-command=>sub{$dlg->destroy})->pack(-side=>'left');
}  
############
sub open_hosts {	# Test existence and then try to read hosts file, on succes invoke nscgi dialog
  my($file)=@_;
  if(-e $file) {
    &read_hosts($file);
    &file_dialog("Open file","Select file to read","nscgi.cfg",\&open_nscgi);
  } else {&file_error($file,0)}
}  
############
sub read_hosts {	# Actual reading of file hosts, parsing into %object hash
  my($file)=@_;
  my($line,$host,$parent);
  undef %object;
  if(-r $file) {
    $dlg->destroy;
    open(HOST,$file) or die("$file : $!");
    while($line=<HOST>) {
      next unless(($host,$parent)=$line=~/^host\[([^\]]+).*?;.*?;([^;]+)?/);
      $object{$host}->{"Parent"}=$parent;
    }
    close(HOST);
  } else {&file_error($file,1)}
}  
############
sub open_nscgi {	# Test existence and then try to read nscgi file
  my($file)=@_;
  if(-e $file) {
    &read_nscgi($file);
  } else {&file_error($file,0)}
}
############
sub read_nscgi {  # Actual nscgi reading, parsing and comparing records in %object, filling @images
  my($file)=@_;
  my($line,$host,$img,$x,$y);
  undef @images;
  my %seen=();
  if(-r $file) {
    $dlg->destroy;
    open(NSCGI,$file) or die("$file : $!");
    while($line=<NSCGI>) {
      next unless(($host,$img,$x,$y)=$line=~/^hostextinfo\[([^\]]+)\]=;([^;]+);.*?;.*?;.*?;(\d+)?,?(\d+)?/);
      if(exists $object{$host}) {
	$object{$host}->{"Img"}=$img;
	unless($seen{$img}++) {
	  push(@images,$img);
	}
      } 
      $object{$host}->{"Coord"}->{"x"}=$x;
      $object{$host}->{"Coord"}->{"y"}=$y;
    }
    close(NSCGI);
    &update_objects;
    &create_objects;
  } else {&file_error($file,1)}
}  
############
sub update_objects { # Correcting coordinates and image names for disapproving records
  my($obj,$x,$y);
  my $counter=0;
  for $obj (keys %object) {
    unless (defined $object{$obj}->{"Coord"}->{"y"}) {
      $counter+=60;
      $x=60+($counter%800);
      $y=(int($counter/800)+1)*40;
      $object{$obj}->{"Coord"}->{"x"}=$x;
      $object{$obj}->{"Coord"}->{"y"}=$y;
    }
    unless (defined $object{$obj}->{Img}) {
      $object{$obj}->{Img}="unknown.gif";
    }
  }
}  
############
sub file_error { # Toplevel error dialog
  my($text,$errid)=@_;
  my @errors=(" : File doesn't exist! ",
              " : Can't read file! ",
              "",
	     );
  my($err,$label,$ef,$but,$map);
  $err=$main->Toplevel(-title=>"Error");
  $err->geometry(&get_geometry);
  $err->transient($main);
  $err->protocol('WM_DELETE_WINDOW',sub{});
  $err->grab;
  $ef=$err->Frame()->pack;
  $map=$ef->Label(-bitmap=>'error')->pack(-side,'left',-padx,5,-pady,5);
  $label=$ef->Label(-text=>$text.$errors[$errid])->pack(-side=>'left');
  $but=$err->Button(-text=>'OK',-command=>sub{$err->destroy;$dlg->grab})
           ->pack(-padx,5,-pady,5);
  $but->focus;	   
}  
############
sub create_objects { # Create canvas objects according to %object, fill in ID's
  my($obj,$img,$lineid,$l,$t,$r,$b);
  for $obj (keys %object) {
    $img=$main->Photo(-file,$imgpath.$object{$obj}->{"Img"});
    $object{$obj}->{"ID"}=$canvas->createImage($object{$obj}->{"Coord"}->{"x"},$object{$obj}->{"Coord"}->{"y"},-image,$img,-tags,["moveable",$obj]);
    ($l,$t,$r,$b)=$canvas->bbox($object{$obj}->{"ID"});
    $object{$obj}->{LabelID}=$canvas->createText($l+20,$b+10,-font,'fixed',-text,$obj);
    if(defined($object{$obj}->{"Parent"}) && (exists $object{$object{$obj}->{"Parent"}})) {
      $lineid=$canvas->createLine($l+(int(($r-$l)/2)),$t+(int(($b-$t)/2)),$object{$object{$obj}->{"Parent"}}->{"Coord"}->{"x"},$object{$object{$obj}->{"Parent"}}->{"Coord"}->{"y"},-tags,["line"]);
      $object{$obj}->{LineID}->{$lineid}=$object{$obj}->{"Parent"};
      $object{$object{$obj}->{"Parent"}}->{LineID}->{$lineid}=$obj;
    }
  }
  $canvas->lower("line","moveable");
  $menubar->entryconfigure('Save',-state=>'normal');
  $menubar->entryconfigure('Open',-state=>'disabled');
}  
############
sub save_state { # Call save file if not exists or user wants to overwrite
  my($file)=@_;
  if((!(-e $file) || ((-e $file) && (&overwrite($file)))) && ($file ne "")) {
    &save_file($file);
  } 
}  
############
sub save_file { # Save %object records into $file
  my($file)=@_;
  my($obj,$img1,$img2);
  unless(open(SAVE,">$file")) {
    &file_error("$file:$!",2);
    return
  }
  $dlg->destroy;
  for $obj (keys %object) {
    ($img1,$img2)=$object{$obj}{Img}=~/(\w+\.)gif/ ? ($1."gif",$1."gd2") : ("",""); 
    print SAVE "hostextinfo[$obj]=;$img1;;$img2;$obj;",$object{$obj}{Coord}{x},",",$object{$obj}{Coord}{y},";;\n";
  }
  close(SAVE);
}  
############
sub overwrite {	# Toplevel overwrite dialog, returns 0 or 1
  my($file)=@_;
  my($over,$map,$text,$butf,$textf,$yes,$no,$retval);
  $over=$main->Toplevel(-title=>'Overwrite file?');
  $over->geometry(&get_geometry);
  $over->transient($main);
  $over->grab;
  $textf=$over->Frame->pack;
  $map=$textf->Label(-bitmap=>'error')->pack(-side,'left',-padx,5,-pady,5);
  $text=$textf->Label(-text=>"File already exists!\nOverwrite?")->pack(-side=>'left');
  $butf=$over->Frame->pack;
  $yes=$butf->Button(-text=>"Yes",-command=>sub{$retval=1})->pack(-side,'left');
  $no=$butf->Button(-text=>"No",-command=>sub{$retval=0})->pack(-side,'left');
  $over->waitVariable(\$retval);
  $over->grabRelease;
  $over->destroy;
  return $retval;
}
############
sub change_type { # Callback for mouse button 3 - change type (image) of the object
  my($id,$name,$img,$choice);
  $id=$canvas->find("withtag","current");
  $name=($canvas->gettags($id))[1];
  $img=$object{$name}{Img};
  $choice=&choose_type($img);
  $object{$name}{Img}=$choice;
  $img=$main->Photo(-file=>$imgpath.$choice);
  $canvas->itemconfigure($id,-image=>$img);
}  
############
sub choose_type { # Change type top level widget - returns image file
  my($current)=@_;
  my($type,$img,$i);
  $type=$main->Toplevel(-title=>"Choose type");
  $type->geometry(&get_geometry);
  $type->transient($main);
  for $img (@images) {
    $i=$main->Photo(-file=>$imgpath.$img); 
    $type->Radiobutton(-image=>$i,
                       -value=>$img,
		       -variable=>\$current,
		       -command=>sub{$type->grabRelease;$type->destroy},
		      )
         ->pack(-anchor=>'w');
  }
  $type->update;
  $type->grab;
  $type->waitWindow;
  return $current;
}  
############
sub quit_it { # Obvious
  $main->destroy;
  exit;
}  
############
